@@ -137,6 +137,8 @@ typedef struct {
   int limit;
 } grn_view;
 
+#define GRN_TABLE_SORT_GEO            (0x02<<0)
+
 #define GRN_OBJ_TMP_OBJECT 0x80000000
 
 #define GRN_DB_OBJP(obj) \
@@ -179,6 +181,22 @@ struct _grn_proc {
   //  grn_obj results[16];
 };
 
+#define GRN_PROC_GET_VAR(name) (grn_proc_get_var(ctx, user_data, name, strlen(name)))
+#define GRN_PROC_GET_VAR_BY_OFFSET(offset) (grn_proc_get_var_by_offset(ctx, user_data, offset))
+#define GRN_PROC_ALLOC(domain, flags) (grn_proc_alloc(ctx, user_data, domain, flags))
+
+grn_obj *grn_proc_get_var(grn_ctx *ctx, grn_user_data *user_data,
+                          const char *name, unsigned name_size);
+
+grn_obj *grn_proc_get_var_by_offset(grn_ctx *ctx, grn_user_data *user_data,
+                                    unsigned int offset);
+
+GRN_API grn_obj *grn_proc_alloc(grn_ctx *ctx, grn_user_data *user_data,
+                                grn_id domain, grn_obj_flags flags);
+
+grn_obj *grn_expr_get_or_add_var(grn_ctx *ctx, grn_obj *expr,
+                                 const char *name, unsigned name_size);
+
 typedef struct _grn_accessor_view grn_accessor_view;
 
 struct _grn_accessor_view {
@@ -234,6 +252,9 @@ int grn_vector_delimit(grn_ctx *ctx, grn_obj *vector);
 int grn_vector_size(grn_ctx *ctx, grn_obj *vector);
 */
 
+unsigned int grn_vector_pop_element(grn_ctx *ctx, grn_obj *vector,
+                                    const char **str, unsigned int *weight, grn_id *domain);
+
 grn_rc grn_vector_delimit(grn_ctx *ctx, grn_obj *v, unsigned int weight, grn_id domain);
 
 grn_rc grn_db_init_builtin_types(grn_ctx *ctx);
@@ -290,7 +311,7 @@ struct _grn_expr {
 GRN_API grn_rc grn_expr_clear_vars(grn_ctx *ctx, grn_obj *expr);
 
 grn_rc grn_expr_parser_close(grn_ctx *ctx);
-grn_rc grn_obj_cast(grn_ctx *ctx, grn_obj *src, grn_obj *dest, int addp);
+GRN_API grn_rc grn_obj_cast(grn_ctx *ctx, grn_obj *src, grn_obj *dest, int addp);
 
 /**
  * grn_table_open:
@@ -398,6 +419,10 @@ grn_id grn_table_at(grn_ctx *ctx, grn_obj *table, grn_id id);
 grn_obj *grn_table_create_for_group(grn_ctx *ctx, const char *name, unsigned name_size,
                                     const char *path, grn_obj_flags flags,
                                     grn_obj *group_key, grn_obj *value_type);
+
+#define KEY_NAME "_key"
+#define ID_NAME "_id"
+
 #ifdef __cplusplus
 }
 #endif
