@@ -1,5 +1,5 @@
 /* -*- c-basic-offset: 2 -*- */
-/* Copyright(C) 2009 Brazil
+/* Copyright(C) 2009-2010 Brazil
 
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
@@ -27,6 +27,7 @@
 #include "util.h"
 #include "output.h"
 #include "pat.h"
+#include "geo.h"
 
 #ifndef O_NOFOLLOW
 #define O_NOFOLLOW 0
@@ -341,9 +342,13 @@ proc_select(grn_ctx *ctx, int nargs, grn_obj **args, grn_user_data *user_data)
 static grn_obj *
 proc_define_selector(grn_ctx *ctx, int nargs, grn_obj **args, grn_user_data *user_data)
 {
-  uint32_t nvars;
+  uint32_t i, nvars;
   grn_expr_var *vars;
   grn_proc_get_info(ctx, user_data, &vars, &nvars, NULL);
+  for (i = 1; i < nvars; i++) {
+    GRN_TEXT_SET(ctx, &((vars + i)->value),
+                 GRN_TEXT_VALUE(VAR(i)), GRN_TEXT_LEN(VAR(i)));
+  }
   grn_proc_create(ctx,
                   GRN_TEXT_VALUE(VAR(0)), GRN_TEXT_LEN(VAR(0)),
                   GRN_PROC_COMMAND, proc_select, NULL, NULL, nvars - 1, vars + 1);
@@ -524,7 +529,6 @@ grn_column_create_flags_to_text(grn_ctx *ctx, grn_obj *buf, grn_obj_flags flags)
   }
   switch (flags & GRN_OBJ_COMPRESS_MASK) {
   case GRN_OBJ_COMPRESS_NONE:
-    GRN_TEXT_PUTS(ctx, buf, "|COMPRESS_NONE");
     break;
   case GRN_OBJ_COMPRESS_ZLIB:
     GRN_TEXT_PUTS(ctx, buf, "|COMPRESS_ZLIB");
@@ -2049,197 +2053,28 @@ func_now(grn_ctx *ctx, int nargs, grn_obj **args, grn_user_data *user_data)
   return obj;
 }
 
-#define GEO_RESOLUTION   3600000
-#define GEO_RADIOUS      6357303
-#define GEO_BES_C1       6334834
-#define GEO_BES_C2       6377397
-#define GEO_BES_C3       0.006674
-#define GEO_GRS_C1       6335439
-#define GEO_GRS_C2       6378137
-#define GEO_GRS_C3       0.006694
-#define GEO_INT2RAD(x)   ((M_PI / (GEO_RESOLUTION * 180)) * x)
-
 static grn_obj *
 func_geo_in_circle(grn_ctx *ctx, int nargs, grn_obj **args, grn_user_data *user_data)
 {
   grn_obj *obj;
   unsigned char r = GRN_FALSE;
   if (nargs == 3) {
-    grn_obj *pos = args[0], *pos1 = args[1], *pos2 = args[2], pos1_, pos2_;
-    grn_id domain = pos->header.domain;
-    if (domain == GRN_DB_TOKYO_GEO_POINT || domain == GRN_DB_WGS84_GEO_POINT) {
-      double lng0, lat0, lng1, lat1, lng2, lat2, x, y, d;
-      if (pos1->header.domain != domain) {
-        GRN_OBJ_INIT(&pos1_, GRN_BULK, 0, domain);
-        if (grn_obj_cast(ctx, pos1, &pos1_, 0)) { goto exit; }
-        pos1 = &pos1_;
-      }
-      lng0 = GEO_INT2RAD(((grn_geo_point *)GRN_BULK_HEAD(pos))->longitude);
-      lat0 = GEO_INT2RAD(((grn_geo_point *)GRN_BULK_HEAD(pos))->latitude);
-      lng1 = GEO_INT2RAD(((grn_geo_point *)GRN_BULK_HEAD(pos1))->longitude);
-      lat1 = GEO_INT2RAD(((grn_geo_point *)GRN_BULK_HEAD(pos1))->latitude);
-      x = (lng1 - lng0) * cos((lat0 + lat1) * 0.5);
-      y = (lat1 - lat0);
-      d = (x * x) + (y * y);
-      switch (pos2->header.domain) {
-      case GRN_DB_INT32 :
-        r = (sqrt(d) * GEO_RADIOUS) <= GRN_INT32_VALUE(pos2);
-        break;
-      case GRN_DB_UINT32 :
-        r = (sqrt(d) * GEO_RADIOUS) <= GRN_UINT32_VALUE(pos2);
-        break;
-      case GRN_DB_INT64 :
-        r = (sqrt(d) * GEO_RADIOUS) <= GRN_INT64_VALUE(pos2);
-        break;
-      case GRN_DB_UINT64 :
-        r = (sqrt(d) * GEO_RADIOUS) <= GRN_UINT64_VALUE(pos2);
-        break;
-      case GRN_DB_FLOAT :
-        r = (sqrt(d) * GEO_RADIOUS) <= GRN_FLOAT_VALUE(pos2);
-        break;
-      case GRN_DB_SHORT_TEXT :
-      case GRN_DB_TEXT :
-      case GRN_DB_LONG_TEXT :
-        GRN_OBJ_INIT(&pos2_, GRN_BULK, 0, domain);
-        if (grn_obj_cast(ctx, pos2, &pos2_, 0)) { goto exit; }
-        pos2 = &pos2_;
-        /* fallthru */
-      case GRN_DB_TOKYO_GEO_POINT :
-      case GRN_DB_WGS84_GEO_POINT :
-        if (domain != pos2->header.domain) { /* todo */ goto exit; }
-        lng2 = GEO_INT2RAD(((grn_geo_point *)GRN_BULK_HEAD(pos2))->longitude);
-        lat2 = GEO_INT2RAD(((grn_geo_point *)GRN_BULK_HEAD(pos2))->latitude);
-        x = (lng2 - lng1) * cos((lat1 + lat2) * 0.5);
-        y = (lat2 - lat1);
-        r = d <= (x * x) + (y * y);
-        break;
-      default :
-        goto exit;
-      }
-    }
+    r = grn_geo_in_circle(ctx, args[0], args[1], args[2]);
   }
-exit :
   if ((obj = GRN_PROC_ALLOC(GRN_DB_UINT32, 0))) {
     GRN_UINT32_SET(ctx, obj, r);
   }
   return obj;
 }
 
-grn_rc
-grn_geo_search(grn_ctx *ctx, grn_obj *obj, grn_obj **args, int nargs,
-               grn_obj *res, grn_operator op)
-{
-  grn_id domain;
-  double lng0, lat0, lng1, lat1, lng2, lat2, x, y, d;
-  grn_obj *proc, *pat, *pos1, *pos2, pos1_, pos2_;
-  if (nargs != 4) { goto exit; }
-  pat = grn_ctx_at(ctx, obj->header.domain);
-  proc = args[0];
-  pos1 = args[2];
-  pos2 = args[3];
-  domain = pat->header.domain;
-  if (domain != GRN_DB_TOKYO_GEO_POINT && domain != GRN_DB_WGS84_GEO_POINT) { goto exit; }
-  if (pos1->header.domain != domain) {
-    GRN_OBJ_INIT(&pos1_, GRN_BULK, 0, domain);
-    if (grn_obj_cast(ctx, pos1, &pos1_, 0)) { goto exit; }
-    pos1 = &pos1_;
-  }
-  lng1 = GEO_INT2RAD(((grn_geo_point *)GRN_BULK_HEAD(pos1))->longitude);
-  lat1 = GEO_INT2RAD(((grn_geo_point *)GRN_BULK_HEAD(pos1))->latitude);
-  switch (pos2->header.domain) {
-  case GRN_DB_INT32 :
-    d = GRN_INT32_VALUE(pos2);
-    d = (d / GEO_RADIOUS) * (d / GEO_RADIOUS);
-    break;
-  case GRN_DB_UINT32 :
-    d = GRN_UINT32_VALUE(pos2);
-    d = (d / GEO_RADIOUS) * (d / GEO_RADIOUS);
-    break;
-  case GRN_DB_INT64 :
-    d = GRN_INT64_VALUE(pos2);
-    d = (d / GEO_RADIOUS) * (d / GEO_RADIOUS);
-    break;
-  case GRN_DB_UINT64 :
-    d = GRN_UINT64_VALUE(pos2);
-    d = (d / GEO_RADIOUS) * (d / GEO_RADIOUS);
-    break;
-  case GRN_DB_FLOAT :
-    d = GRN_FLOAT_VALUE(pos2);
-    d = (d / GEO_RADIOUS) * (d / GEO_RADIOUS);
-    break;
-  case GRN_DB_SHORT_TEXT :
-  case GRN_DB_TEXT :
-  case GRN_DB_LONG_TEXT :
-    GRN_OBJ_INIT(&pos2_, GRN_BULK, 0, domain);
-    if (grn_obj_cast(ctx, pos2, &pos2_, 0)) { goto exit; }
-    pos2 = &pos2_;
-    /* fallthru */
-  case GRN_DB_TOKYO_GEO_POINT :
-  case GRN_DB_WGS84_GEO_POINT :
-    if (domain != pos2->header.domain) { /* todo */ goto exit; }
-    lng2 = GEO_INT2RAD(((grn_geo_point *)GRN_BULK_HEAD(pos2))->longitude);
-    lat2 = GEO_INT2RAD(((grn_geo_point *)GRN_BULK_HEAD(pos2))->latitude);
-    x = (lng2 - lng1) * cos((lat1 + lat2) * 0.5);
-    y = (lat2 - lat1);
-    d = ((x * x) + (y * y));
-    break;
-  default :
-    goto exit;
-  }
-  {
-    uint32_t s, h;
-    grn_id tid;
-    grn_geo_point pos;
-    grn_table_cursor *tc = grn_table_cursor_open(ctx, pat, NULL, 0,
-                                                 GRN_BULK_HEAD(pos1),
-                                                 sizeof(grn_geo_point),
-                                                 0, -1, GRN_CURSOR_PREFIX);
-    for (s = 0, h = 256; s <= h * 16 && (tid = grn_table_cursor_next(ctx, tc)); s++) {
-      grn_table_get_key(ctx, pat, tid, &pos, sizeof(grn_geo_point));
-      lng0 = GEO_INT2RAD(pos.longitude);
-      lat0 = GEO_INT2RAD(pos.latitude);
-      x = (lng1 - lng0) * cos((lat0 + lat1) * 0.5);
-      y = (lat1 - lat0);
-      if (((x * x) + (y * y)) <= d) {
-        grn_ii_at(ctx, (grn_ii *)obj, tid, (grn_hash *)res, op);
-        h++;
-      }
-    }
-    grn_table_cursor_close(ctx, tc);
-  }
-exit :
-  grn_ii_resolve_sel_and(ctx, (grn_hash *)res, op);
-  return ctx->rc;
-}
-
 static grn_obj *
 func_geo_in_rectangle(grn_ctx *ctx, int nargs, grn_obj **args, grn_user_data *user_data)
 {
   grn_obj *obj;
   unsigned char r = GRN_FALSE;
   if (nargs == 3) {
-    grn_obj *pos = args[0], *pos1 = args[1], *pos2 = args[2], pos1_, pos2_;
-    grn_geo_point *p, *p1, *p2;
-    grn_id domain = pos->header.domain;
-    if (domain == GRN_DB_TOKYO_GEO_POINT || domain == GRN_DB_WGS84_GEO_POINT) {
-      if (pos1->header.domain != domain) {
-        GRN_OBJ_INIT(&pos1_, GRN_BULK, 0, domain);
-        if (grn_obj_cast(ctx, pos1, &pos1_, 0)) { goto exit; }
-        pos1 = &pos1_;
-      }
-      if (pos2->header.domain != domain) {
-        GRN_OBJ_INIT(&pos2_, GRN_BULK, 0, domain);
-        if (grn_obj_cast(ctx, pos2, &pos2_, 0)) { goto exit; }
-        pos2 = &pos2_;
-      }
-      p = ((grn_geo_point *)GRN_BULK_HEAD(pos));
-      p1 = ((grn_geo_point *)GRN_BULK_HEAD(pos1));
-      p2 = ((grn_geo_point *)GRN_BULK_HEAD(pos2));
-      r = ((p1->longitude <= p->longitude) && (p->longitude <= p2->longitude) &&
-           (p1->latitude <= p->latitude) && (p->latitude <= p2->latitude));
-    }
+    r = grn_geo_in_rectangle(ctx, args[0], args[1], args[2]);
   }
-exit :
   if ((obj = GRN_PROC_ALLOC(GRN_DB_UINT32, 0))) {
     GRN_UINT32_SET(ctx, obj, r);
   }
@@ -2252,25 +2087,8 @@ func_geo_distance(grn_ctx *ctx, int nargs, grn_obj **args, grn_user_data *user_d
   grn_obj *obj;
   double d = 0;
   if (nargs == 2) {
-    grn_obj *pos = args[0], *pos1 = args[1], pos1_;
-    grn_id domain = pos->header.domain;
-    if (domain == GRN_DB_TOKYO_GEO_POINT || domain == GRN_DB_WGS84_GEO_POINT) {
-      double lng0, lat0, lng1, lat1, x, y;
-      if (pos1->header.domain != domain) {
-        GRN_OBJ_INIT(&pos1_, GRN_BULK, 0, domain);
-        if (grn_obj_cast(ctx, pos1, &pos1_, 0)) { goto exit; }
-        pos1 = &pos1_;
-      }
-      lng0 = GEO_INT2RAD(((grn_geo_point *)GRN_BULK_HEAD(pos))->longitude);
-      lat0 = GEO_INT2RAD(((grn_geo_point *)GRN_BULK_HEAD(pos))->latitude);
-      lng1 = GEO_INT2RAD(((grn_geo_point *)GRN_BULK_HEAD(pos1))->longitude);
-      lat1 = GEO_INT2RAD(((grn_geo_point *)GRN_BULK_HEAD(pos1))->latitude);
-      x = (lng1 - lng0) * cos((lat0 + lat1) * 0.5);
-      y = (lat1 - lat0);
-      d = sqrt((x * x) + (y * y)) * GEO_RADIOUS;
-    }
+    d = grn_geo_distance(ctx, args[0], args[1]);
   }
-exit :
   if ((obj = GRN_PROC_ALLOC(GRN_DB_FLOAT, 0))) {
     GRN_FLOAT_SET(ctx, obj, d);
   }
@@ -2283,25 +2101,8 @@ func_geo_distance2(grn_ctx *ctx, int nargs, grn_obj **args, grn_user_data *user_
   grn_obj *obj;
   double d = 0;
   if (nargs == 2) {
-    grn_obj *pos = args[0], *pos1 = args[1], pos1_;
-    grn_id domain = pos->header.domain;
-    if (domain == GRN_DB_TOKYO_GEO_POINT || domain == GRN_DB_WGS84_GEO_POINT) {
-      double lng0, lat0, lng1, lat1, x, y;
-      if (pos1->header.domain != domain) {
-        GRN_OBJ_INIT(&pos1_, GRN_BULK, 0, domain);
-        if (grn_obj_cast(ctx, pos1, &pos1_, 0)) { goto exit; }
-        pos1 = &pos1_;
-      }
-      lng0 = GEO_INT2RAD(((grn_geo_point *)GRN_BULK_HEAD(pos))->longitude);
-      lat0 = GEO_INT2RAD(((grn_geo_point *)GRN_BULK_HEAD(pos))->latitude);
-      lng1 = GEO_INT2RAD(((grn_geo_point *)GRN_BULK_HEAD(pos1))->longitude);
-      lat1 = GEO_INT2RAD(((grn_geo_point *)GRN_BULK_HEAD(pos1))->latitude);
-      x = sin(fabs(lng1 - lng0) * 0.5);
-      y = sin(fabs(lat1 - lat0) * 0.5);
-      d = asin(sqrt((y * y) + cos(lat0) * cos(lat1) * x * x)) * 2 * GEO_RADIOUS;
-    }
+    d = grn_geo_distance2(ctx, args[0], args[1]);
   }
-exit :
   if ((obj = GRN_PROC_ALLOC(GRN_DB_FLOAT, 0))) {
     GRN_FLOAT_SET(ctx, obj, d);
   }
@@ -2314,59 +2115,8 @@ func_geo_distance3(grn_ctx *ctx, int nargs, grn_obj **args, grn_user_data *user_
   grn_obj *obj;
   double d = 0;
   if (nargs == 2) {
-    grn_obj *pos = args[0], *pos1 = args[1], pos1_;
-    grn_id domain = pos->header.domain;
-    switch (domain) {
-    case GRN_DB_TOKYO_GEO_POINT :
-      {
-        double lng0, lat0, lng1, lat1, p, q, r, m, n, x, y;
-        if (pos1->header.domain != domain) {
-          GRN_OBJ_INIT(&pos1_, GRN_BULK, 0, domain);
-          if (grn_obj_cast(ctx, pos1, &pos1_, 0)) { goto exit; }
-          pos1 = &pos1_;
-        }
-        lng0 = GEO_INT2RAD(((grn_geo_point *)GRN_BULK_HEAD(pos))->longitude);
-        lat0 = GEO_INT2RAD(((grn_geo_point *)GRN_BULK_HEAD(pos))->latitude);
-        lng1 = GEO_INT2RAD(((grn_geo_point *)GRN_BULK_HEAD(pos1))->longitude);
-        lat1 = GEO_INT2RAD(((grn_geo_point *)GRN_BULK_HEAD(pos1))->latitude);
-        p = (lat0 + lat1) * 0.5;
-        q = (1 - GEO_BES_C3 * sin(p) * sin(p));
-        r = sqrt(q);
-        m = GEO_BES_C1 / (q * r);
-        n = GEO_BES_C2 / r;
-        x = n * cos(p) * fabs(lng0 - lng1);
-        y = m * fabs(lat0 - lat1);
-        d = sqrt((x * x) + (y * y));
-      }
-      break;
-    case  GRN_DB_WGS84_GEO_POINT :
-      {
-        double lng0, lat0, lng1, lat1, p, q, r, m, n, x, y;
-        if (pos1->header.domain != domain) {
-          GRN_OBJ_INIT(&pos1_, GRN_BULK, 0, domain);
-          if (grn_obj_cast(ctx, pos1, &pos1_, 0)) { goto exit; }
-          pos1 = &pos1_;
-        }
-        lng0 = GEO_INT2RAD(((grn_geo_point *)GRN_BULK_HEAD(pos))->longitude);
-        lat0 = GEO_INT2RAD(((grn_geo_point *)GRN_BULK_HEAD(pos))->latitude);
-        lng1 = GEO_INT2RAD(((grn_geo_point *)GRN_BULK_HEAD(pos1))->longitude);
-        lat1 = GEO_INT2RAD(((grn_geo_point *)GRN_BULK_HEAD(pos1))->latitude);
-        p = (lat0 + lat1) * 0.5;
-        q = (1 - GEO_GRS_C3 * sin(p) * sin(p));
-        r = sqrt(q);
-        m = GEO_GRS_C1 / (q * r);
-        n = GEO_GRS_C2 / r;
-        x = n * cos(p) * fabs(lng0 - lng1);
-        y = m * fabs(lat0 - lat1);
-        d = sqrt((x * x) + (y * y));
-      }
-      break;
-    default :
-      /* todo */
-      break;
-    }
+    d = grn_geo_distance3(ctx, args[0], args[1]);
   }
-exit :
   if ((obj = GRN_PROC_ALLOC(GRN_DB_FLOAT, 0))) {
     GRN_FLOAT_SET(ctx, obj, d);
   }
