@@ -31,8 +31,7 @@ static grn_obj *database, *comments, *content, *expression, *variable, *result;
 void
 cut_startup(void)
 {
-  tmp_directory = g_build_filename(grn_test_get_base_dir(),
-                                   "tmp",
+  tmp_directory = g_build_filename(grn_test_get_tmp_dir(),
                                    "test-table-select-normalize",
                                    NULL);
   database_path = g_build_filename(tmp_directory, "database.groonga", NULL);
