@@ -31,7 +31,8 @@ class HTTPTest < Test::Unit::TestCase
   def test_status
     response = get(command_path(:status))
     assert_equal("application/json", response.content_type)
-    assert_equal(["alloc_count", "starttime", "uptime", "version"],
+    assert_equal(["alloc_count", "cache_hit_rate", "n_queries",
+                  "starttime", "uptime", "version"],
                  JSON.parse(response.body)[1].keys.sort)
   end
 
@@ -51,7 +52,7 @@ class HTTPTest < Test::Unit::TestCase
                             :content_type => "application/json")
     @groonga_pid = nil
 
-    assert_raise(Errno::ECONNREFUSED) do
+    assert_raise(Errno::ECONNREFUSED, EOFError) do
       get(command_path(:shutdown))
     end
   end
